from typing import Type, Tuple, List, Dict, Union, Optional, ClassVar
from os import listdir, remove
from os.path import isfile, join, exists
from sys import version_info
from json import loads, dumps, JSONDecodeError
from datetime import datetime
from functools import partial
from dataclasses import dataclass, Field, field
from setuptools import find_packages

GenericAlias = type(List)


PACKAGE_INFO_FILE = "PACKAGE.json"
PACKAGE_INFO_SOURCES = dict(
    name="PACKAGE_NAME",
    version="VERSION",
    long_description="README*",
    license="LICENSE*"
)
REQUIREMENTS_SOURCE = "requirements.txt"
EXTRAS_REQUIREMENTS_SOURCE = "{extra}-requirements.txt"
EXCLUDE_TESTS = True

PYTHON_VERSIONS_MAJOR = ["2", "3"]
PYTHON_VERSIONS_MINOR = [
    "2.3", "2.4", "2.5", "2.6", "2.7",
    "3.0", "3.1", "3.2", "3.3", "3.4", "3.5", "3.6", "3.7", "3.8", "3.9"
]
CLASSIFIERS_DEVELOPMENT_STATUS = [
    "Development Status :: 1 - Planning",
    "Development Status :: 2 - Pre-Alpha",
    "Development Status :: 3 - Alpha",
    "Development Status :: 4 - Beta",
    "Development Status :: 5 - Production/Stable",
    "Development Status :: 6 - Mature",
    "Development Status :: 7 - Inactive",
]
DEVELOPMENT_STATUS_LEVEL = dict(
    zip(
        CLASSIFIERS_DEVELOPMENT_STATUS,
        ("plan", "prealpha", "alpha", "beta", None, None, None)
    )
)
DEVELOPMENT_STATUS_MAP = {
    str(i+1): status
    for i, status in enumerate(CLASSIFIERS_DEVELOPMENT_STATUS)
}
DEVELOPMENT_STATUS_MAP.update(
    **{
        status.partition(" :: ")[2].lower(): status
        for status in CLASSIFIERS_DEVELOPMENT_STATUS
    },
    **{
        status.partition(" - ")[2].lower(): status
        for status in CLASSIFIERS_DEVELOPMENT_STATUS
    },
    **{
        "design": DEVELOPMENT_STATUS_MAP["1"],
        "dev": DEVELOPMENT_STATUS_MAP["2"],
        "development": DEVELOPMENT_STATUS_MAP["2"],
        "test": DEVELOPMENT_STATUS_MAP["3"],
        "testing": DEVELOPMENT_STATUS_MAP["3"],
        "pre-release": DEVELOPMENT_STATUS_MAP["4"],
        "staging": DEVELOPMENT_STATUS_MAP["4"],
        "release": DEVELOPMENT_STATUS_MAP["5"],
        "production": DEVELOPMENT_STATUS_MAP["5"],
        "stable": DEVELOPMENT_STATUS_MAP["5"],
    }
)
CLASSIFIERS_LICENSE = """License :: Aladdin Free Public License (AFPL)
License :: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication
License :: DFSG approved
License :: Eiffel Forum License (EFL)
License :: Free For Educational Use
License :: Free For Home Use
License :: Free for non-commercial use
License :: Freely Distributable
License :: Free To Use But Restricted
License :: Freeware
License :: Netscape Public License (NPL)
License :: Nokia Open Source License (NOKOS)
License :: OSI Approved
License :: OSI Approved :: Academic Free License (AFL)
License :: OSI Approved :: Apache Software License
License :: OSI Approved :: Apple Public Source License
License :: OSI Approved :: Artistic License
License :: OSI Approved :: Attribution Assurance License
License :: OSI Approved :: BSD License
License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)
License :: OSI Approved :: Common Public License
License :: OSI Approved :: Eiffel Forum License
License :: OSI Approved :: European Union Public Licence 1.0 (EUPL 1.0)
License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)
License :: OSI Approved :: GNU Affero General Public License v3
License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)
License :: OSI Approved :: GNU Free Documentation License (FDL)
License :: OSI Approved :: GNU General Public License (GPL)
License :: OSI Approved :: GNU General Public License v2 (GPLv2)
License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)
License :: OSI Approved :: GNU General Public License v3 (GPLv3)
License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)
License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)
License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)
License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)
License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)
License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)
License :: OSI Approved :: IBM Public License
License :: OSI Approved :: Intel Open Source License
License :: OSI Approved :: ISC License (ISCL)
License :: OSI Approved :: Jabber Open Source License
License :: OSI Approved :: MIT License
License :: OSI Approved :: MITRE Collaborative Virtual Workspace License (CVW)
License :: OSI Approved :: Motosoto License
License :: OSI Approved :: Mozilla Public License 1.0 (MPL)
License :: OSI Approved :: Mozilla Public License 1.1 (MPL 1.1)
License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)
License :: OSI Approved :: Nethack General Public License
License :: OSI Approved :: Nokia Open Source License
License :: OSI Approved :: Open Group Test Suite License
License :: OSI Approved :: Python License (CNRI Python License)
License :: OSI Approved :: Python Software Foundation License
License :: OSI Approved :: Qt Public License (QPL)
License :: OSI Approved :: Ricoh Source Code Public License
License :: OSI Approved :: Sleepycat License
License :: OSI Approved :: Sun Industry Standards Source License (SISSL)
License :: OSI Approved :: Sun Public License
License :: OSI Approved :: University of Illinois/NCSA Open Source License
License :: OSI Approved :: Vovida Software License 1.0
License :: OSI Approved :: W3C License
License :: OSI Approved :: X.Net License
License :: OSI Approved :: zlib/libpng License
License :: OSI Approved :: Zope Public License
License :: Other/Proprietary License
License :: Public Domain
License :: Repoze Public License""".split("\n")
CLASSIFIERS = {
    "Operating System": """Operating System :: Android
Operating System :: BeOS
Operating System :: iOS
Operating System :: MacOS
Operating System :: MacOS :: MacOS 9
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft
Operating System :: Microsoft :: MS-DOS
Operating System :: Microsoft :: Windows
Operating System :: Microsoft :: Windows :: Windows 3.1 or Earlier
Operating System :: Microsoft :: Windows :: Windows 7
Operating System :: Microsoft :: Windows :: Windows 95/98/2000
Operating System :: Microsoft :: Windows :: Windows CE
Operating System :: Microsoft :: Windows :: Windows NT/2000
Operating System :: Microsoft :: Windows :: Windows Server 2003
Operating System :: Microsoft :: Windows :: Windows Server 2008
Operating System :: Microsoft :: Windows :: Windows Vista
Operating System :: Microsoft :: Windows :: Windows XP
Operating System :: OS/2
Operating System :: OS Independent
Operating System :: Other OS
Operating System :: PalmOS
Operating System :: PDA Systems
Operating System :: POSIX
Operating System :: POSIX :: AIX
Operating System :: POSIX :: BSD
Operating System :: POSIX :: BSD :: BSD/OS
Operating System :: POSIX :: BSD :: FreeBSD
Operating System :: POSIX :: BSD :: NetBSD
Operating System :: POSIX :: BSD :: OpenBSD
Operating System :: POSIX :: GNU Hurd
Operating System :: POSIX :: HP-UX
Operating System :: POSIX :: IRIX
Operating System :: POSIX :: Linux
Operating System :: POSIX :: Other
Operating System :: POSIX :: SCO
Operating System :: POSIX :: SunOS/Solaris
Operating System :: Unix""".split("\n"),
    "Environment": """Environment :: Console
Environment :: Console :: Curses
Environment :: Console :: Framebuffer
Environment :: Console :: Newt
Environment :: Console :: svgalib
Environment :: Handhelds/PDA's
Environment :: MacOS X
Environment :: MacOS X :: Aqua
Environment :: MacOS X :: Carbon
Environment :: MacOS X :: Cocoa
Environment :: No Input/Output (Daemon)
Environment :: OpenStack
Environment :: Other Environment
Environment :: Plugins
Environment :: Web Environment
Environment :: Web Environment :: Buffet
Environment :: Web Environment :: Mozilla
Environment :: Web Environment :: ToscaWidgets
Environment :: Win32 (MS Windows)
Environment :: X11 Applications
Environment :: X11 Applications :: Gnome
Environment :: X11 Applications :: GTK
Environment :: X11 Applications :: KDE
Environment :: X11 Applications :: Qt""".split("\n"),
    "Framework": """Framework :: BFG
Framework :: Bob
Framework :: Bottle
Framework :: Buildout
Framework :: Buildout :: Extension
Framework :: Buildout :: Recipe
Framework :: Chandler
Framework :: CherryPy
Framework :: CubicWeb
Framework :: Django
Framework :: Django :: 1.4
Framework :: Django :: 1.5
Framework :: Django :: 1.6
Framework :: Django :: 1.7
Framework :: Django :: 1.8
Framework :: Django :: 1.9
Framework :: Flask
Framework :: IDLE
Framework :: IPython
Framework :: Opps
Framework :: Paste
Framework :: Plone
Framework :: Plone :: 3.2
Framework :: Plone :: 3.3
Framework :: Plone :: 4.0
Framework :: Plone :: 4.1
Framework :: Plone :: 4.2
Framework :: Plone :: 4.3
Framework :: Plone :: 5.0
Framework :: Pylons
Framework :: Pyramid
Framework :: Pytest
Framework :: Review Board
Framework :: Robot Framework
Framework :: Scrapy
Framework :: Setuptools Plugin
Framework :: Sphinx
Framework :: Sphinx :: Extension
Framework :: Sphinx :: Theme
Framework :: Trac
Framework :: Tryton
Framework :: TurboGears
Framework :: TurboGears :: Applications
Framework :: TurboGears :: Widgets
Framework :: Twisted
Framework :: ZODB
Framework :: Zope2
Framework :: Zope3""".split("\n"),
    "Natural Language": """Natural Language :: Afrikaans
Natural Language :: Arabic
Natural Language :: Bengali
Natural Language :: Bosnian
Natural Language :: Bulgarian
Natural Language :: Cantonese
Natural Language :: Catalan
Natural Language :: Chinese (Simplified)
Natural Language :: Chinese (Traditional)
Natural Language :: Croatian
Natural Language :: Czech
Natural Language :: Danish
Natural Language :: Dutch
Natural Language :: English
Natural Language :: Esperanto
Natural Language :: Finnish
Natural Language :: French
Natural Language :: Galician
Natural Language :: German
Natural Language :: Greek
Natural Language :: Hebrew
Natural Language :: Hindi
Natural Language :: Hungarian
Natural Language :: Icelandic
Natural Language :: Indonesian
Natural Language :: Italian
Natural Language :: Japanese
Natural Language :: Javanese
Natural Language :: Korean
Natural Language :: Latin
Natural Language :: Latvian
Natural Language :: Macedonian
Natural Language :: Malay
Natural Language :: Marathi
Natural Language :: Norwegian
Natural Language :: Panjabi
Natural Language :: Persian
Natural Language :: Polish
Natural Language :: Portuguese
Natural Language :: Portuguese (Brazilian)
Natural Language :: Romanian
Natural Language :: Russian
Natural Language :: Serbian
Natural Language :: Slovak
Natural Language :: Slovenian
Natural Language :: Spanish
Natural Language :: Swedish
Natural Language :: Tamil
Natural Language :: Telugu
Natural Language :: Thai
Natural Language :: Turkish
Natural Language :: Ukranian
Natural Language :: Urdu
Natural Language :: Vietnamese""".split("\n"),
    "Intended Audience": """Intended Audience :: Customer Service
Intended Audience :: Developers
Intended Audience :: Education
Intended Audience :: End Users/Desktop
Intended Audience :: Financial and Insurance Industry
Intended Audience :: Healthcare Industry
Intended Audience :: Information Technology
Intended Audience :: Legal Industry
Intended Audience :: Manufacturing
Intended Audience :: Other Audience
Intended Audience :: Religion
Intended Audience :: Science/Research
Intended Audience :: System Administrators
Intended Audience :: Telecommunications Industry""".split("\n"),
    "Topic": """Topic :: Adaptive Technologies
Topic :: Artistic Software
Topic :: Communications
Topic :: Communications :: BBS
Topic :: Communications :: Chat
Topic :: Communications :: Chat :: AOL Instant Messenger
Topic :: Communications :: Chat :: ICQ
Topic :: Communications :: Chat :: Internet Relay Chat
Topic :: Communications :: Chat :: Unix Talk
Topic :: Communications :: Conferencing
Topic :: Communications :: Email
Topic :: Communications :: Email :: Address Book
Topic :: Communications :: Email :: Email Clients (MUA)
Topic :: Communications :: Email :: Filters
Topic :: Communications :: Email :: Mailing List Servers
Topic :: Communications :: Email :: Mail Transport Agents
Topic :: Communications :: Email :: Post-Office
Topic :: Communications :: Email :: Post-Office :: IMAP
Topic :: Communications :: Email :: Post-Office :: POP3
Topic :: Communications :: Fax
Topic :: Communications :: FIDO
Topic :: Communications :: File Sharing
Topic :: Communications :: File Sharing :: Gnutella
Topic :: Communications :: File Sharing :: Napster
Topic :: Communications :: Ham Radio
Topic :: Communications :: Internet Phone
Topic :: Communications :: Telephony
Topic :: Communications :: Usenet News
Topic :: Database
Topic :: Database :: Database Engines/Servers
Topic :: Database :: Front-Ends
Topic :: Desktop Environment
Topic :: Desktop Environment :: File Managers
Topic :: Desktop Environment :: Gnome
Topic :: Desktop Environment :: GNUstep
Topic :: Desktop Environment :: K Desktop Environment (KDE)
Topic :: Desktop Environment :: K Desktop Environment (KDE) :: Themes
Topic :: Desktop Environment :: PicoGUI
Topic :: Desktop Environment :: PicoGUI :: Applications
Topic :: Desktop Environment :: PicoGUI :: Themes
Topic :: Desktop Environment :: Screen Savers
Topic :: Desktop Environment :: Window Managers
Topic :: Desktop Environment :: Window Managers :: Afterstep
Topic :: Desktop Environment :: Window Managers :: Afterstep :: Themes
Topic :: Desktop Environment :: Window Managers :: Applets
Topic :: Desktop Environment :: Window Managers :: Blackbox
Topic :: Desktop Environment :: Window Managers :: Blackbox :: Themes
Topic :: Desktop Environment :: Window Managers :: CTWM
Topic :: Desktop Environment :: Window Managers :: CTWM :: Themes
Topic :: Desktop Environment :: Window Managers :: Enlightenment
Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Epplets
Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Themes DR15
Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Themes DR16
Topic :: Desktop Environment :: Window Managers :: Enlightenment :: Themes DR17
Topic :: Desktop Environment :: Window Managers :: Fluxbox
Topic :: Desktop Environment :: Window Managers :: Fluxbox :: Themes
Topic :: Desktop Environment :: Window Managers :: FVWM
Topic :: Desktop Environment :: Window Managers :: FVWM :: Themes
Topic :: Desktop Environment :: Window Managers :: IceWM
Topic :: Desktop Environment :: Window Managers :: IceWM :: Themes
Topic :: Desktop Environment :: Window Managers :: MetaCity
Topic :: Desktop Environment :: Window Managers :: MetaCity :: Themes
Topic :: Desktop Environment :: Window Managers :: Oroborus
Topic :: Desktop Environment :: Window Managers :: Oroborus :: Themes
Topic :: Desktop Environment :: Window Managers :: Sawfish
Topic :: Desktop Environment :: Window Managers :: Sawfish :: Themes 0.30
Topic :: Desktop Environment :: Window Managers :: Sawfish :: Themes pre-0.30
Topic :: Desktop Environment :: Window Managers :: Waimea
Topic :: Desktop Environment :: Window Managers :: Waimea :: Themes
Topic :: Desktop Environment :: Window Managers :: Window Maker
Topic :: Desktop Environment :: Window Managers :: Window Maker :: Applets
Topic :: Desktop Environment :: Window Managers :: Window Maker :: Themes
Topic :: Desktop Environment :: Window Managers :: XFCE
Topic :: Desktop Environment :: Window Managers :: XFCE :: Themes
Topic :: Documentation
Topic :: Documentation :: Sphinx
Topic :: Education
Topic :: Education :: Computer Aided Instruction (CAI)
Topic :: Education :: Testing
Topic :: Games/Entertainment
Topic :: Games/Entertainment :: Arcade
Topic :: Games/Entertainment :: Board Games
Topic :: Games/Entertainment :: First Person Shooters
Topic :: Games/Entertainment :: Fortune Cookies
Topic :: Games/Entertainment :: Multi-User Dungeons (MUD)
Topic :: Games/Entertainment :: Puzzle Games
Topic :: Games/Entertainment :: Real Time Strategy
Topic :: Games/Entertainment :: Role-Playing
Topic :: Games/Entertainment :: Side-Scrolling/Arcade Games
Topic :: Games/Entertainment :: Simulation
Topic :: Games/Entertainment :: Turn Based Strategy
Topic :: Home Automation
Topic :: Internet
Topic :: Internet :: File Transfer Protocol (FTP)
Topic :: Internet :: Finger
Topic :: Internet :: Log Analysis
Topic :: Internet :: Name Service (DNS)
Topic :: Internet :: Proxy Servers
Topic :: Internet :: WAP
Topic :: Internet :: WWW/HTTP
Topic :: Internet :: WWW/HTTP :: Browsers
Topic :: Internet :: WWW/HTTP :: Dynamic Content
Topic :: Internet :: WWW/HTTP :: Dynamic Content :: CGI Tools/Libraries
Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Message Boards
Topic :: Internet :: WWW/HTTP :: Dynamic Content :: News/Diary
Topic :: Internet :: WWW/HTTP :: Dynamic Content :: Page Counters
Topic :: Internet :: WWW/HTTP :: HTTP Servers
Topic :: Internet :: WWW/HTTP :: Indexing/Search
Topic :: Internet :: WWW/HTTP :: Session
Topic :: Internet :: WWW/HTTP :: Site Management
Topic :: Internet :: WWW/HTTP :: Site Management :: Link Checking
Topic :: Internet :: WWW/HTTP :: WSGI
Topic :: Internet :: WWW/HTTP :: WSGI :: Application
Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware
Topic :: Internet :: WWW/HTTP :: WSGI :: Server
Topic :: Internet :: Z39.50
Topic :: Multimedia
Topic :: Multimedia :: Graphics
Topic :: Multimedia :: Graphics :: 3D Modeling
Topic :: Multimedia :: Graphics :: 3D Rendering
Topic :: Multimedia :: Graphics :: Capture
Topic :: Multimedia :: Graphics :: Capture :: Digital Camera
Topic :: Multimedia :: Graphics :: Capture :: Scanners
Topic :: Multimedia :: Graphics :: Capture :: Screen Capture
Topic :: Multimedia :: Graphics :: Editors
Topic :: Multimedia :: Graphics :: Editors :: Raster-Based
Topic :: Multimedia :: Graphics :: Editors :: Vector-Based
Topic :: Multimedia :: Graphics :: Graphics Conversion
Topic :: Multimedia :: Graphics :: Presentation
Topic :: Multimedia :: Graphics :: Viewers
Topic :: Multimedia :: Sound/Audio
Topic :: Multimedia :: Sound/Audio :: Analysis
Topic :: Multimedia :: Sound/Audio :: Capture/Recording
Topic :: Multimedia :: Sound/Audio :: CD Audio
Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Playing
Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Ripping
Topic :: Multimedia :: Sound/Audio :: CD Audio :: CD Writing
Topic :: Multimedia :: Sound/Audio :: Conversion
Topic :: Multimedia :: Sound/Audio :: Editors
Topic :: Multimedia :: Sound/Audio :: MIDI
Topic :: Multimedia :: Sound/Audio :: Mixers
Topic :: Multimedia :: Sound/Audio :: Players
Topic :: Multimedia :: Sound/Audio :: Players :: MP3
Topic :: Multimedia :: Sound/Audio :: Sound Synthesis
Topic :: Multimedia :: Sound/Audio :: Speech
Topic :: Multimedia :: Video
Topic :: Multimedia :: Video :: Capture
Topic :: Multimedia :: Video :: Conversion
Topic :: Multimedia :: Video :: Display
Topic :: Multimedia :: Video :: Non-Linear Editor
Topic :: Office/Business
Topic :: Office/Business :: Financial
Topic :: Office/Business :: Financial :: Accounting
Topic :: Office/Business :: Financial :: Investment
Topic :: Office/Business :: Financial :: Point-Of-Sale
Topic :: Office/Business :: Financial :: Spreadsheet
Topic :: Office/Business :: Groupware
Topic :: Office/Business :: News/Diary
Topic :: Office/Business :: Office Suites
Topic :: Office/Business :: Scheduling
Topic :: Other/Nonlisted Topic
Topic :: Printing
Topic :: Religion
Topic :: Scientific/Engineering
Topic :: Scientific/Engineering :: Artificial Intelligence
Topic :: Scientific/Engineering :: Artificial Life
Topic :: Scientific/Engineering :: Astronomy
Topic :: Scientific/Engineering :: Atmospheric Science
Topic :: Scientific/Engineering :: Bio-Informatics
Topic :: Scientific/Engineering :: Chemistry
Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)
Topic :: Scientific/Engineering :: GIS
Topic :: Scientific/Engineering :: Human Machine Interfaces
Topic :: Scientific/Engineering :: Image Recognition
Topic :: Scientific/Engineering :: Information Analysis
Topic :: Scientific/Engineering :: Interface Engine/Protocol Translator
Topic :: Scientific/Engineering :: Mathematics
Topic :: Scientific/Engineering :: Medical Science Apps.
Topic :: Scientific/Engineering :: Physics
Topic :: Scientific/Engineering :: Visualization
Topic :: Security
Topic :: Security :: Cryptography
Topic :: Sociology
Topic :: Sociology :: Genealogy
Topic :: Sociology :: History
Topic :: Software Development
Topic :: Software Development :: Assemblers
Topic :: Software Development :: Bug Tracking
Topic :: Software Development :: Build Tools
Topic :: Software Development :: Code Generators
Topic :: Software Development :: Compilers
Topic :: Software Development :: Debuggers
Topic :: Software Development :: Disassemblers
Topic :: Software Development :: Documentation
Topic :: Software Development :: Embedded Systems
Topic :: Software Development :: Internationalization
Topic :: Software Development :: Interpreters
Topic :: Software Development :: Libraries
Topic :: Software Development :: Libraries :: Application Frameworks
Topic :: Software Development :: Libraries :: Java Libraries
Topic :: Software Development :: Libraries :: Perl Modules
Topic :: Software Development :: Libraries :: PHP Classes
Topic :: Software Development :: Libraries :: Pike Modules
Topic :: Software Development :: Libraries :: pygame
Topic :: Software Development :: Libraries :: Python Modules
Topic :: Software Development :: Libraries :: Ruby Modules
Topic :: Software Development :: Libraries :: Tcl Extensions
Topic :: Software Development :: Localization
Topic :: Software Development :: Object Brokering
Topic :: Software Development :: Object Brokering :: CORBA
Topic :: Software Development :: Pre-processors
Topic :: Software Development :: Quality Assurance
Topic :: Software Development :: Testing
Topic :: Software Development :: Testing :: Traffic Generation
Topic :: Software Development :: User Interfaces
Topic :: Software Development :: Version Control
Topic :: Software Development :: Version Control :: CVS
Topic :: Software Development :: Version Control :: RCS
Topic :: Software Development :: Version Control :: SCCS
Topic :: Software Development :: Widget Sets
Topic :: System
Topic :: System :: Archiving
Topic :: System :: Archiving :: Backup
Topic :: System :: Archiving :: Compression
Topic :: System :: Archiving :: Mirroring
Topic :: System :: Archiving :: Packaging
Topic :: System :: Benchmark
Topic :: System :: Boot
Topic :: System :: Boot :: Init
Topic :: System :: Clustering
Topic :: System :: Console Fonts
Topic :: System :: Distributed Computing
Topic :: System :: Emulators
Topic :: System :: Filesystems
Topic :: System :: Hardware
Topic :: System :: Hardware :: Hardware Drivers
Topic :: System :: Hardware :: Mainframes
Topic :: System :: Hardware :: Symmetric Multi-processing
Topic :: System :: Installation/Setup
Topic :: System :: Logging
Topic :: System :: Monitoring
Topic :: System :: Networking
Topic :: System :: Networking :: Firewalls
Topic :: System :: Networking :: Monitoring
Topic :: System :: Networking :: Monitoring :: Hardware Watchdog
Topic :: System :: Networking :: Time Synchronization
Topic :: System :: Operating System
Topic :: System :: Operating System Kernels
Topic :: System :: Operating System Kernels :: BSD
Topic :: System :: Operating System Kernels :: GNU Hurd
Topic :: System :: Operating System Kernels :: Linux
Topic :: System :: Power (UPS)
Topic :: System :: Recovery Tools
Topic :: System :: Shells
Topic :: System :: Software Distribution
Topic :: System :: Systems Administration
Topic :: System :: Systems Administration :: Authentication/Directory
Topic :: System :: Systems Administration :: Authentication/Directory :: LDAP
Topic :: System :: Systems Administration :: Authentication/Directory :: NIS
Topic :: System :: System Shells
Topic :: Terminals
Topic :: Terminals :: Serial
Topic :: Terminals :: Telnet
Topic :: Terminals :: Terminal Emulators/X Terminals
Topic :: Text Editors
Topic :: Text Editors :: Documentation
Topic :: Text Editors :: Emacs
Topic :: Text Editors :: Integrated Development Environments (IDE)
Topic :: Text Editors :: Text Processing
Topic :: Text Editors :: Word Processors
Topic :: Text Processing
Topic :: Text Processing :: Filters
Topic :: Text Processing :: Fonts
Topic :: Text Processing :: General
Topic :: Text Processing :: Indexing
Topic :: Text Processing :: Linguistic
Topic :: Text Processing :: Markup
Topic :: Text Processing :: Markup :: HTML
Topic :: Text Processing :: Markup :: LaTeX
Topic :: Text Processing :: Markup :: SGML
Topic :: Text Processing :: Markup :: VRML
Topic :: Text Processing :: Markup :: XML
Topic :: Utilities""".split("\n")
}


Attribute = Union[
    str,
    List[str],
    Dict[str, str],
    Dict[str, List[str]],
    Dict[str, Dict[str, List[str]]]
]
AttributeType = Union[Type[Attribute], GenericAlias]


def make_setup(package_path: str, overwrite=False):
    def _transform_value(expected_type: AttributeType, value_to_transform: str) -> Attribute:
        origin = str if expected_type is Union[int, str] else getattr(expected_type, "__origin__", expected_type)
        if origin is str:
            return origin(value_to_transform)
        if origin is list:
            items_type = expected_type.__args__[0]
            return [_transform_value(items_type, item.strip()) for item in value_to_transform.split(",")]
        if origin is dict:
            values_type = expected_type.__args__[1]
            items = {
                item[0].strip(): item[2].strip()
                for item in (
                    kv.partition(":")
                    for kv in value_to_transform.split(";")
                )
            }
            return {
                key: _transform_value(values_type, item.strip())
                for key, item in items.items()
            }
        raise NotImplementedError(expected_type)

    def _load_file(file_name: str) -> Tuple[str, Optional[str]]:
        if file_name.endswith("*"):
            paths = [
                name for name in listdir(package_path)
                if (
                    name.startswith(file_name[:-1]) and
                    isfile(join(package_path, name))
                )
            ]
            if not paths:
                return "", None
            if len(paths) > 1:
                raise OSError(
                    f"\nFound multiple files matching {file_name!r} in {package_path}:"
                    f"\n    {dumps(paths, indent=4)[1:-1]}"
                )
            file_name = paths.pop()
        if not (exists(file_name) and isfile(file_name)):
            return "", None
        with open(file_name) as f:
            return f.read().strip(), file_name

    def _load_data(file_name: str) -> Dict:
        file_data, file_name = _load_file(file_name)
        if not file_data:
            return {}
        file_data = {k.lower().replace(" ", "_"): v for k, v in loads(file_data).items()}
        output = {}
        for name in PackageData.__dataclass_fields__:
            if name == "package_path":
                continue
            keys = [name, *PackageData.ALTERNATIVE_FIELDS.get(name, ())]
            for k in keys:
                if k in file_data:
                    if name in output:
                        desc = ", ".join(map(repr, keys))
                        raise ValueError(f"Specified multiple options for {desc}")
                    if isinstance(file_data[k], str):
                        output[name] = _transform_value(PackageData.__dataclass_fields__[name].type, file_data[k])
                    else:
                        output[name] = file_data[k]
        return output

    def _load_key(
        file_name: str,
        for_key: str,
        old_value: Union[str, list, dict, None]
    ) -> Union[str, list, dict, None]:
        value_data, file_name = _load_file(file_name)
        if not value_data:
            return old_value
        value_for_key = _transform_value(PackageData.__dataclass_fields__[for_key].type, value_data)
        if old_value is not None and value_for_key != old_value:
            raise ValueError(
                f"\nInitialized with following value for {for_key!r}:\n    {old_value!r}"
                f"\n\nReceived differing value from {join(package_path, file_name)!r}:\n    {value_for_key!r}"
            )
        return value_for_key

    setup_path = join(package_path, "setup.py")
    if overwrite and exists(setup_path):
        remove(setup_path)
    config_path = join(package_path, "setup.cfg")
    if overwrite and exists(config_path):
        remove(config_path)
    data = dict(package_path=package_path)
    data.update(_load_data(PACKAGE_INFO_FILE))
    _, license_file = _load_file(PACKAGE_INFO_SOURCES["license"])
    if license_file:
        data.update(license_file=license_file)
    _, description_file = _load_file(PACKAGE_INFO_SOURCES["long_description"])
    if description_file:
        data.update(description_file=description_file)
    for key, path in PACKAGE_INFO_SOURCES.items():
        new_value = _load_key(path, key, data.get(key))
        if new_value is None and key not in ("license", ):
            raise ValueError(f"Missing value for {key!r}")
        data[key] = new_value
    PackageData(**data).write_script()


@dataclass
class PackageData:
    __dataclass_fields__: ClassVar[Dict[str, Field]]

    IGNORE_IF_EMPTY = (
        "tests_require",
        "extras_require",
        "scripts",
        "entry_points",
        "keywords"
    )
    ALTERNATIVE_FIELDS = dict(
        python_requires=("python", ),
        version=("ver", "v"),
        author_email=("email", ),
        project_urls=("urls", ),
        exclude_packages=("exclude", ),
        development_status=("status", ),
        operating_systems=("operating_system", "os"),
        environments=("environment", ),
        frameworks=("framework", ),
        intended_audiences=("intended_audience", "audiences", "audience"),
        topics=("topic",),
        natural_languages=("natural_language",),
    )

    package_path: str
    name: str
    version: str
    url: str = None

    keywords: str = None
    description: str = None
    project_urls: Dict[str, str] = field(default_factory=dict)

    author: str = None
    author_email: str = None
    maintainer: str = None
    maintainer_email: str = None

    first_release: str = None  # ToDo
    long_description: str = None
    long_description_content_type = "text/markdown"
    description_file: str = None
    license: str = None  # ToDo
    license_file: str = None  # ToDo

    python_requires: str = f"~={version_info.major}.{version_info.minor}"
    extras_require: Dict[str, List[str]] = field(default_factory=dict)
    tests_require: List[str] = field(default_factory=list)

    exclude_packages: List[str] = field(default_factory=lambda: ["tests*", "*.tests*"])
    scripts: List[str] = field(default_factory=list)
    entry_points: Dict[str, Dict[str, List[str]]] = field(default_factory=dict)

    development_status: Union[int, str] = None
    classifier_license: str = None
    operating_systems: List[str] = None
    environments: List[str] = None
    frameworks: List[str] = None

    intended_audiences: List[str] = None
    topics: List[str] = None
    natural_languages: List[str] = None

    @property
    def packages(self):
        return find_packages(
            self.package_path,
            exclude=[f"{self.name}.{item}" for item in self.exclude_packages]
        )

    @property
    def setup_script(self) -> str:
        def _make(key: str, result: str, *, level: int, for_dict: bool, replace_lb=True):
            spaces = " " * level * 4
            if replace_lb:
                result = result.strip().replace("\n", f"\n{spaces}")
            key = f"\"{key}\": " if for_dict else f"{key}="
            return f"\n{spaces}{key}{result}"

        def make_string(key: str, value: str = None, *, level: int = 1, for_dict=False):
            if "\n" not in value:
                return _make(key, f"\"{value}\"", level=level, for_dict=for_dict)
            return _make(key, f"r\"\"\"{''.join(value)}\n\"\"\"", level=level, for_dict=for_dict, replace_lb=False)

        def make_list(key: str, value: List[str], *, level: int = 1, for_dict=False):
            if not value:
                return _make(key, "[]", level=level, for_dict=for_dict)
            value = [f"\n    \"{x.strip()}\"" for x in sorted(value)]
            return _make(key, f"[{','.join(value)}\n]", level=level, for_dict=for_dict)

        def make_dict(key: str, value: Dict, *, level: int = 1):
            if not value:
                return _make(key, "{}", level=level, for_dict=False)
            values_type = self.__dataclass_fields__[key].type.__args__[1]
            items = sorted(value.items(), key=lambda kv: kv[0])
            if values_type is str:
                values = (
                    make_string(k, v, level=level, for_dict=True)
                    for k, v in items
                )
            elif values_type is List[str]:
                values = (
                    make_list(k, v, level=level, for_dict=True)
                    for k, v in items
                )
            else:
                assert values_type.__origin__ is dict, f"{key}: {value!r}"
                assert values_type.__args__[1] is List[str], f"{key}: {value!r}"
                values = (
                    f"\"{k}\": " + "\n".join(
                        make_list(kk, vv, level=level+1, for_dict=True)
                        for kk, vv in sorted(v.items(), key=lambda kv: kv[0])
                    )
                    for k, v in items
                )
            return _make(key, f"{{{','.join(values)}\n}}", level=level, for_dict=False)

        def make(key: str, value: Attribute = None, *, level: int = 1):
            value = value if value is not None else getattr(self, key)
            if value is None or key in self.IGNORE_IF_EMPTY and not value:
                return None
            if isinstance(value, str):
                return make_string(key, value.strip(), level=level)
            if isinstance(value, list):
                return make_list(key, value, level=level)
            if isinstance(value, dict):
                return make_dict(key, value, level=level)
            raise TypeError(value)

        attributes = filter(
            lambda x: x is not None,
            map(
                make,
                (
                    "name",
                    "version",
                    "url",
                    "project_urls",
                    "author",
                    "author_email",
                    "maintainer",
                    "maintainer_email",
                    "license",
                    "license_file",
                    "keywords",
                    "classifiers",
                    "description",
                    "long_description_content_type",
                    "long_description",
                    "packages",
                    "python_requires",
                    "install_requires",
                    "tests_require",
                    "extras_require",
                    "scripts",
                    "entry_points",
                )
            )
        )

        return (
            f"\"\"\"Setup script for {self.name} package.\n\n"
            f"Generated using {__package__} on {str(datetime.now()).rpartition('.')[0]}.\n"
            f"\"\"\"\n\n"
            f"from setuptools import setup\n\n"
            f"setup({','.join(attributes)}\n)\n"
        )

    @property
    def setup_config(self):
        return (
            "[metadata]\n"
            f"description-file = {self.description_file}\n"
            f"version = {self.version}\n"
            f"long_description_content_type = {self.long_description_content_type}\n" +
            f"license_files = {self.license_file}\n" if self.license_file else ""
        )

    def write_script(self):
        file_path = join(self.package_path, "setup.py")
        if exists(file_path):
            raise FileExistsError(file_path)
        with open(file_path, "w") as file:
            file.write(self.setup_script)
        file_path = join(self.package_path, "setup.cfg")
        if exists(file_path):
            raise FileExistsError(file_path)
        with open(file_path, "w") as file:
            file.write(self.setup_config)

    def _parse_status(self):
        if self.development_status is None:
            return
        status = self.development_status.lower()
        if status.startswith("development status"):
            status = status[len("development status"):].strip().strip(":").strip(":").strip()
        try:
            self.classifiers.append(DEVELOPMENT_STATUS_MAP[status])
            version_postfix = DEVELOPMENT_STATUS_LEVEL[DEVELOPMENT_STATUS_MAP[status]]
            if version_postfix:
                self.version = f"{self.version}.{version_postfix}"
        except KeyError:
            available = {k: [] for k in CLASSIFIERS_DEVELOPMENT_STATUS}
            for k, v in DEVELOPMENT_STATUS_MAP.items():
                try:
                    available[v].append(str(int(k)))
                except ValueError:
                    pass
                available[v].append(f"\"{k}\"")
            available = "\n".join(f"\"{k}\", also as:\n    {', '.join(v)}" for k, v in available.items())
            raise ValueError(
                f"Unknown Development Status {self.development_status!r}, "
                f"available statuses:\n    {available}"
            )

    def _parse_license(self):
        if self.license is None and self.classifier_license is None:
            self.classifier_license = "Other/Proprietary License"
            release = self.version.split(".")[0]
            if self.first_release:
                initial_release = self.first_release.split(".")[0]
                if initial_release != release:
                    release = f"{initial_release}-{release}"
            self.license = f"Proprietary License. Copyright (c) {release} {self.author}. All Rights Reserved."

        if self.classifier_license is not None:
            if not self.classifier_license.lower().startswith("license ::"):
                self.classifier_license = f"License :: {self.classifier_license.strip().title()}"
            self.classifier_license = self.classifier_license.strip().title()
            if self.classifier_license not in CLASSIFIERS_LICENSE:
                available = "\n    ".join(CLASSIFIERS_LICENSE)
                raise ValueError(f"Unknown license classifier, available classifiers:\n    {available}")
            split = self.classifier_license.split(" :: ")
            if len(split) > 2:
                assert len(split) == 3, self.classifier_license
                self.classifiers.append(f"{split[0]} :: {split[1]}")
            self.classifiers.append(self.classifier_license)
        elif self.license is not None:
            first_row = self.license.strip().split("\n")[0].strip()
            self.license = first_row
            possible_licenses = [item for item in CLASSIFIERS_LICENSE if first_row in item]
            if possible_licenses:
                if len(possible_licenses) > 1:
                    raise ValueError(
                        f"Multiple options possible for license classifier: "
                        f"{', '.join(map(repr, possible_licenses))}"
                    )
                self.classifier_license = possible_licenses.pop()
                self.classifiers.append(self.classifier_license)

    def _parse_classifiers(self, label: str):
        all_classifiers = CLASSIFIERS[label]
        possible_items = {
            key.lower(): key
            for key in all_classifiers
        }
        available = "\n    ".join(all_classifiers)
        items = sorted(getattr(self, f"{label.lower().replace(' ', '_')}s") or [])
        for item in items:
            if not item.lower().startswith(f"{label.lower()} ::"):
                item = f"{label} :: {item.strip()}"
            try:
                item = possible_items[item.strip().lower()]
            except KeyError:
                raise ValueError(f"Unknown {label.lower()} classifier, available classifiers:\n    {available}")
            split = item.split(" :: ")
            item = split.pop(0)
            while split:
                item = f"{item} :: {split.pop(0)}"
                if item not in self.classifiers:
                    self.classifiers.append(item)

    def _parse_python(self):
        major_versions, minor_versions, exclude_versions = [], [], []
        for requirement in [x.strip() for x in self.python_requires.split(",")]:
            if any(
                requirement.startswith(x)
                for x in ("~=", "==", "!=", "<=", ">=")
            ):
                prefix, requirement = requirement[:2], requirement[2:].strip()
            elif requirement.startswith("<") or requirement.startswith(">"):
                prefix, requirement = requirement[:1], requirement[1:].strip()
            else:
                raise ValueError(f"Can't parse requirement {requirement!r}")
            major, *minor = requirement.split(".")
            version = major if not minor else f"{major}.{minor.pop(0)}"
            if prefix == "!=":
                if version not in exclude_versions:
                    exclude_versions.append(version)
                continue

            if prefix == ">=":
                majors = [x for x in PYTHON_VERSIONS_MAJOR if x >= major]
                versions = [x for x in PYTHON_VERSIONS_MINOR if x >= version]
            elif prefix == "<=":
                majors = [x for x in PYTHON_VERSIONS_MAJOR if x <= major]
                versions = [x for x in PYTHON_VERSIONS_MINOR if x <= version]
            elif prefix == ">":
                majors = [x for x in PYTHON_VERSIONS_MAJOR if x > major]
                versions = [x for x in PYTHON_VERSIONS_MINOR if x > version]
            elif prefix == "<":
                majors = [x for x in PYTHON_VERSIONS_MAJOR if x < major]
                versions = [x for x in PYTHON_VERSIONS_MINOR if x < version]
            elif prefix == "~=":
                major_max = str(int(major)+1)
                majors = [x for x in PYTHON_VERSIONS_MAJOR if major_max > x >= major]
                versions = [x for x in PYTHON_VERSIONS_MINOR if major_max > x >= version]
            else:
                assert prefix == "==", prefix
                majors, versions = [major], [version]

            if version == major:
                versions = []

            for major in majors:
                if major not in major_versions:
                    major_versions.append(major)
            for version in versions:
                if version not in minor_versions:
                    minor_versions.append(version)
        major_versions = sorted(x for x in major_versions if x not in exclude_versions)
        self.classifiers.append("Programming Language :: Python")
        for major in major_versions:
            only = " :: Only" if len(major_versions) == 1 else ""
            self.classifiers.append(f"Programming Language :: Python :: {major}{only}")
            for version in sorted(x for x in minor_versions if x not in exclude_versions and x.startswith(major)):
                self.classifiers.append(f"Programming Language :: Python :: {version}")

    def _parse_requirements(self):
        def _get_name(requirement: str):
            for char in "!<>=":
                string = requirement.partition(char)[0]
            return string

        def _update_requirements(*, source: str, to_update: Optional[List[str]]) -> List[str]:
            to_update = [] if not to_update else [*to_update]
            existing_names = list(map(_get_name, to_update))
            try:
                with open(source) as file:
                    lines = filter(None, map(lambda l: l.strip(), file.readlines()))
            except FileNotFoundError:
                return to_update
            for requirement in lines:
                name = _get_name(requirement)
                if name not in existing_names:
                    to_update.append(requirement)
            return to_update

        self.install_requires = _update_requirements(
            source=REQUIREMENTS_SOURCE,
            to_update=None
        )
        for extra in {*self.extras_require, "tests", "dev"}:
            self.extras_require[extra] = _update_requirements(
                source=EXTRAS_REQUIREMENTS_SOURCE.format(extra=extra),
                to_update=self.extras_require.get(extra)
            )

        tests_require, self.tests_require = [*self.tests_require], [f"{self.name}[tests]"]
        for item in tests_require:
            if item not in self.extras_require["tests"]:
                self.extras_require["tests"].append(item)

        dev_requires, self.extras_require["dev"] = [*self.extras_require["dev"]], [f"{self.name}[tests]"]
        for item in dev_requires:
            if item not in self.extras_require["tests"]:
                self.extras_require["dev"].append(item)

        if not self.extras_require["tests"]:
            self.tests_require = []
            assert not self.extras_require.pop("tests")
            assert self.extras_require["dev"].pop(0) == f"{self.name}[tests]"

        if not self.extras_require["dev"]:
            self.extras_require.pop("dev")

    def __post_init__(self):
        self.classifiers = []
        self._parse_license()
        self._parse_python()
        self._parse_requirements()
        self._parse_status()
        for label in CLASSIFIERS:
            self._parse_classifiers(label)
